import itertools
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Dict, Optional

from model.animals import random_animal_name
from model.card import Card
from model.deck import Deck
from model.hand import Hand


@dataclass
class Announce:
    bet: Optional[Hand] = None

    @property
    def menteur(self):
        return not self.bet


class Player:
    def __init__(self, name: str = None):
        if not name:
            name = random_animal_name()
        self.name: str = name
        self.hand: Hand = Hand()

    def give(self, card: Card):
        self.hand.give(card)

    def announce(self, current_bet: Optional[Hand]) -> Announce:
        """ A naive player that only trusts what they sees:
        bets his hand, or menteur if his hand is lower. """
        if not current_bet or self.hand.value() > current_bet.value():
            return Announce(self.hand)
        else:
            return Announce()


class Game:
    def __init__(self, players=None, deck: Deck = Deck()):
        if players is None:
            players = []

        self.deck: Deck = deck
        self.players: List[Player] = players
        self.defeats: Dict[Player, int] = defaultdict(lambda: 0)

    @property
    def global_hand(self) -> Hand:
        return Hand([c for p in self.players for c in p.hand])

    def new_game(self):
        self.deck.reset()

        while len(self.players) > 1:
            loser = self.new_turn()

            if self.defeats[loser] == 5:
                print(f"Player {loser} is eliminated!")
                self.players.remove(loser)
            else:
                print(f"Player {loser} lost the round, now playing with {self.defeats[loser] + 1} cards!")

        winner = self.players[0]
        print(f"Game over - Player {winner.name} wins with {len(winner.hand)} cards!")

    def new_turn(self) -> Player:
        """
        Runs a turn of the game.

        :return: the player that lost this turn.
        """
        # Distribution
        self.deck.reset()
        for current_player in self.players:
            current_player.give(self.deck.random_card())

        # Tour
        current_bet: Optional[Hand] = None
        last_player = None
        for current_player in itertools.cycle(self.players):
            print(f"|Player {current_player.name}'s turn >")
            if not current_bet:  # First player, has to bet something
                while not current_bet:
                    announce = current_player.announce(current_bet)
                    if announce.bet:
                        current_bet = announce.bet
                        print(f"Player {current_player.name} starts the round: {current_bet}")
            else:  # Next player, announce or menteur
                announce = current_player.announce(current_bet)
                if announce.bet:
                    while announce.bet.value() < current_bet.value(): # Bad announce!
                        print(f"Invalid bet, {announce.bet} < {current_bet}!")
                        announce = current_player.announce(current_bet)
                    current_bet = announce.bet
                    print(f"Player {current_player.name} bets {current_bet}.")
                else:  # Menteur! Who lost the round?
                    menteur = self.is_menteur(current_bet)
                    print(f"Player {current_player.name} says Menteur... {'Bravo!' if menteur else 'FAIL!'}")
                    loser = last_player if menteur else current_player
                    print(f"Player {loser.name} lost the round!")
                    self.count_defeat(loser)
                    return loser
            last_player = current_player

        # TODO: Put next first player first of list

    def is_menteur(self, bet: Hand):
        to_scan = [Card(c.value) for c in self.global_hand.cards.copy()]  # Keep only values
        to_find = [Card(c.value) for c in bet.cards]
        for card in to_find:
            if card in to_scan:
                to_scan.remove(card)
                continue
            else:
                print(f"Missing a {card}!")
                print(f"Didn't find {bet} in {self.global_hand}: MENTEUR!")
                return False  #

    def count_defeat(self, loser):
        self.defeats[loser] += 1
